@@ -1,7 +1,7 @@
 /*
- * wmnd - window maker network devices - drivers.c
- * drivers definitions - implementation
- * Copyright(c) 2001-2004 by wave++ "Yuri D'Elia" <wavexx@users.sf.net>
+ * drivers: drivers definitions - implementation
+ * Copyright(c) 2001-2008 by wave++ "Yuri D'Elia" <wavexx@users.sf.net>
+ * Distributed under GNU GPL (v2 or above) WITHOUT ANY WARRANTY.
  */
 
 /* local headers */
@@ -9,38 +9,33 @@
 #include "drivers.h"
 #include "messages.h"
 
-/*
- * _list: allocate new elements in the list and assign a name for each device
- * _init: callled for each device allocated by *_list. Prepare device status
- * _getstats: obtain infos about the device
- * _term: eventually deallocate drvdata and close fds
- */
 
-/* helper functions */
+/* common helper functions */
 struct Devices*
 devices_append(struct Devices* list, struct Devices* src)
 {
   list->next = src;
   src->next = NULL;
-
   return src;
 }
 
+
 #ifdef USE_SOLARIS_FPPPD
 #undef drName
 #define drName USE_SOLARIS_FPPPD
 
 /*
  * this driver should work on any system that could compile and execute pppd
- * for linux/solaris. Had require some hacking and has some drawbacks but
- * finally works. always detect at least one device since you can't detect the
- * number of futher connections. The _get function also required a bit of hack
- * to manage correctly the status of the device without messing the graph
+ * for linux/solaris.
  */
 
 /* some needed headers */
 #include <sys/stropts.h>
+#ifdef HAVE_NET_PPP_DEFS_H
+#include <net/ppp_defs.h>
+#else
 #include "ppp_defs.h"
+#endif
 #include "pppio.h"
 #include <errno.h>
 
@@ -78,7 +73,7 @@ solaris_fpppd_strioctl(int fd, int cmd, char* ptr, int ilen, int olen)
     return -1;
   if(str.ic_len != olen)
     msg_drInfo(drName, "strioctl expected %d bytes, got %d for cmd %x",
-        olen, str.ic_len, cmd);
+	olen, str.ic_len, cmd);
   return 0;
 }
 
@@ -101,9 +96,9 @@ solaris_fpppd_init(struct Devices* dev)
     {
       if(drvdata->mustconnect == sfpppd_untested)
       {
-        /* supposing the device is down */
-        drvdata->mustconnect = sfpppd_opened;
-        return 0;
+	/* supposing the device is down */
+	drvdata->mustconnect = sfpppd_opened;
+	return 0;
       }
       drvdata->mustconnect = sfpppd_opened;
     }
@@ -113,7 +108,7 @@ solaris_fpppd_init(struct Devices* dev)
   if(drvdata->mustconnect < sfpppd_connected)
   {
     if(solaris_fpppd_strioctl(drvdata->fd, PPPIO_ATTACH,
-        (char*)&drvdata->unit, sizeof(int), 0) < 0)
+	(char*)&drvdata->unit, sizeof(int), 0) < 0)
     {
       drvdata->mustconnect = sfpppd_opened;
       dev->devstart = 0;
@@ -132,7 +127,7 @@ solaris_fpppd_init(struct Devices* dev)
 void
 solaris_fpppd_term(struct Devices* dev)
 {
-  close(((struct solaris_fpppd_drvdata*) dev->drvdata)->fd);
+  close(((struct solaris_fpppd_drvdata*)dev->drvdata)->fd);
   free(dev->drvdata);
 }
 
@@ -145,14 +140,14 @@ solaris_fpppd_list(const char* devname, struct Devices* list)
   struct solaris_fpppd_drvdata* ndata;
   struct Devices* ptr;
   int dta;
-  
+
   /* device name testing */
   if(devname)
     devn = strdup(devname);
-  
-  if(devn) /* device specified */	
+
+  if(devn) /* device specified */
   {
-    if (sscanf(devn, "ppp%d", &unit) != 1)
+    if(sscanf(devn, "ppp%d", &unit) != 1)
     {
       msg_drInfo(drName, "invalid specified interface '%s'", devn);
       free(devn);
@@ -177,8 +172,8 @@ solaris_fpppd_list(const char* devname, struct Devices* list)
       solaris_fpppd_term(ndev);
 
       if(dta)
-        /* device initialization failed */
-        break; 
+	/* device initialization failed */
+	break;
 
       msg_drInfo(drName, "detected ppp%d",unit);
       unit++;
@@ -187,11 +182,11 @@ solaris_fpppd_list(const char* devname, struct Devices* list)
 
     if(!unit)
       return 0;
-      
+
     dta = unit;
     unit = 0;
   }
-  
+
   /* allocate new dta devices structures */
   ptr = list;
   while(dta--)
@@ -223,17 +218,17 @@ solaris_fpppd_get(struct Devices* dev, unsigned long* ip,
   static struct ppp_stats curp;
   struct solaris_fpppd_drvdata* drvdata = dev->drvdata;
 
-  /* clear vars (not clearing will fuck up all scales when disconnecting */
+  /* clear vars (not clearing will fuck up all scales when disconnecting) */
   *ip = *op = *ib = *ob = 0;
 
   if(drvdata->mustconnect < sfpppd_connected)
   {
     /* trying to activate the device */
-    if (solaris_fpppd_init(dev))
+    if(solaris_fpppd_init(dev))
       return 1;
   }
-  if(solaris_fpppd_strioctl(drvdata->fd, PPPIO_GETSTAT, (char* )&curp,
-      0, sizeof(curp)) < 0)
+  if(solaris_fpppd_strioctl(drvdata->fd, PPPIO_GETSTAT,
+	 (char*)&curp, 0, sizeof(curp)) < 0)
   {
     /* the connection has probably shutted down */
     drvdata->mustconnect = sfpppd_failed;
@@ -297,26 +292,25 @@ solaris_kstat_list(const char* devname, struct Devices* list)
     {
       kstat_read(solaris_kstat_kc, ksp, NULL);
       if((!devname || (devname && !strcmp(devname,ksp->ks_name))) &&
-          kstat_data_lookup(ksp, "link_up") &&
-          kstat_data_lookup(ksp, "ipackets") &&
-          kstat_data_lookup(ksp, "opackets") &&
-          kstat_data_lookup(ksp, "rbytes") &&
-          kstat_data_lookup(ksp, "obytes"))
+	  kstat_data_lookup(ksp, "ipackets") &&
+	  kstat_data_lookup(ksp, "opackets") &&
+	  kstat_data_lookup(ksp, "rbytes") &&
+	  kstat_data_lookup(ksp, "obytes"))
       {
-        /* device is suitable */
-        ndev = (struct Devices*)malloc(sizeof(struct Devices));
-        ndev->devstart = 0;
-        ndev->name = strdup(ksp->ks_name);
-        ndata = (struct solaris_kstat_drvdata*)
-          malloc(sizeof(struct solaris_kstat_drvdata));
-        ndata->ksp = ksp;
-        ptr->next = ndev;
-        ndev->next = NULL;
-        ndev->drvdata = ndata;
-        ptr = ndev;
+	/* device is suitable */
+	ndev = (struct Devices*)malloc(sizeof(struct Devices));
+	ndev->devstart = 0;
+	ndev->name = strdup(ksp->ks_name);
+	ndata = (struct solaris_kstat_drvdata*)
+	  malloc(sizeof(struct solaris_kstat_drvdata));
+	ndata->ksp = ksp;
+	ptr->next = ndev;
+	ndev->next = NULL;
+	ndev->drvdata = ndata;
+	ptr = ndev;
 
-        msg_drInfo(drName, "detected %s",ndev->name);
-        dta++;
+	msg_drInfo(drName, "detected %s",ndev->name);
+	dta++;
       }
     }
   }
@@ -331,12 +325,14 @@ solaris_kstat_init(struct Devices* dev)
     (struct solaris_kstat_drvdata*)dev->drvdata;
 
   drvdata->active = (kstat_named_t*)kstat_data_lookup(drvdata->ksp,"link_up");
+  if(!drvdata->active)
+    msg_drInfo(drName, "device %s can't properly detect link status", dev->name);
   drvdata->in_pkt = (kstat_named_t*)kstat_data_lookup(drvdata->ksp,"ipackets");
   drvdata->in_byte = (kstat_named_t*)kstat_data_lookup(drvdata->ksp,"rbytes");
   drvdata->out_pkt = (kstat_named_t*)kstat_data_lookup(drvdata->ksp,"opackets");
   drvdata->out_byte = (kstat_named_t*)kstat_data_lookup(drvdata->ksp,"obytes");
 
-  return !(drvdata->active && drvdata->in_pkt && drvdata->in_byte &&
+  return !(drvdata->in_pkt && drvdata->in_byte &&
       drvdata->out_pkt && drvdata->out_byte);
 }
 
@@ -348,17 +344,25 @@ solaris_kstat_get(struct Devices* dev, unsigned long* ip, unsigned long* op,
     (struct solaris_kstat_drvdata*)dev->drvdata;
   *ip = *op = *ib = *ob = 0;
 
+  /* a bit ugly, but use the correct type on either 32 or 64bit builds
+     without having to check data_type dynamically or truncate. */
+#if (SIZEOF_UNSIGNED_LONG < 8)
+#define KSTAT_WORD_TYPE ui32
+#else
+#define KSTAT_WORD_TYPE ui64
+#endif
+
   if(kstat_read(solaris_kstat_kc, drvdata->ksp, NULL) == -1)
     return 1;
   else
   {
-    *ip = (drvdata->in_pkt->value.ui32);
-    *op = (drvdata->out_pkt->value.ui32);
-    *ib = (drvdata->in_byte->value.ui32);
-    *ob = (drvdata->out_byte->value.ui32);
+    *ip = (drvdata->in_pkt->value.KSTAT_WORD_TYPE);
+    *op = (drvdata->out_pkt->value.KSTAT_WORD_TYPE);
+    *ib = (drvdata->in_byte->value.KSTAT_WORD_TYPE);
+    *ob = (drvdata->out_byte->value.KSTAT_WORD_TYPE);
   }
 
-  return !drvdata->active->value.ui32;
+  return (drvdata->active? !drvdata->active->value.KSTAT_WORD_TYPE: 0);
 }
 
 void
@@ -374,14 +378,11 @@ solaris_kstat_term(struct Devices* dev)
 #undef drName
 #define drName USE_TESTING_DUMMY
 
-#ifdef USE_SINE_TESTING_DUMMY
 #include <math.h>
 
 /* we need the double of PI, not defined everywhere */
 #undef M_2PI
-#define M_2PI M_PI * 2
-
-#endif
+#define M_2PI (M_PI * 2)
 
 int
 testing_dummy_list(const char* devname, struct Devices* list)
@@ -410,7 +411,6 @@ int
 testing_dummy_get(struct Devices* dev, unsigned long* ip, unsigned long* op,
     unsigned long* ib, unsigned long* ob)
 {
-#ifdef USE_SINE_TESTING_DUMMY
   /* some more fun when debugging! */
   static float v = 0.;
   static unsigned long is = 0;
@@ -418,20 +418,12 @@ testing_dummy_get(struct Devices* dev, unsigned long* ip, unsigned long* op,
 
   *ip = *ib = (is += (16384. * cos(v)) + 16384);
   *op = *ob = (os += (16384. * sin(v)) + 16384);
+  if(v >= M_2PI) v = fmodf(v, M_2PI);
   v += 0.05;
 
-  if(v >= M_PI * 2)
-    v = 0;
-#else
-  *ip = *op = *ib = *ob = 0;
-#endif
   return 1;
 }
 
-void
-testing_dummy_term(struct Devices* dev)
-{}
-
 #endif /* USE_TESTING_DUMMY */
 
 
@@ -447,6 +439,7 @@ testing_dummy_term(struct Devices* dev)
  */
 
 const char* linux_proc_netDevice = "/proc/net/dev";
+const char* linux_proc_tokens = " :\t\n"; /* parse tokens */
 
 int
 linux_proc_list(const char* devname, struct Devices* list)
@@ -454,7 +447,6 @@ linux_proc_list(const char* devname, struct Devices* list)
   FILE* fd;
   int dta = 0;
   char temp[MAXBUF]; /* string buffer */
-  char* tokens = " :\t\n"; /* parse tokens */
   struct Devices* ndev;
   char* p;
 
@@ -487,10 +479,10 @@ linux_proc_list(const char* devname, struct Devices* list)
     while(fgets(temp, MAXBUF, fd))
     {
       /* grab all active devices, adding them as we go */
-      p = strtok(temp, tokens);
+      p = strtok(temp, linux_proc_tokens);
       if(!strncmp(p, "dummy", 5) ||
-          !strncmp(p, "irda", 4) || !strncmp(p, "lo", 3))
-        continue;
+	  !strncmp(p, "irda", 4) || !strncmp(p, "lo", 3))
+	continue;
 
       dta++;
       ndev = (struct Devices*)malloc(sizeof(struct Devices));
@@ -506,13 +498,6 @@ linux_proc_list(const char* devname, struct Devices* list)
   return dta;
 }
 
-int
-linux_proc_init(struct Devices* dev)
-{
-  /* linux proc doesn't need single device initialization */
-  return 0;
-}
-
 int
 linux_proc_get(struct Devices* dev, unsigned long* ip, unsigned long* op,
     unsigned long* ib, unsigned long* ob)
@@ -532,27 +517,24 @@ linux_proc_get(struct Devices* dev, unsigned long* ip, unsigned long* op,
   *ib = *ob = *ip = *op = 0;
 
   while(fgets(temp, MAXBUF, fp))
-    if(strstr(temp, dev->name))
+  {
+    if(!strcmp(strtok(temp, linux_proc_tokens), dev->name))
     {
-      p = strchr(temp, ':');
-      ++p;
+      p = strchr(temp, 0) + 1;
       sscanf(p, "%lu %lu %*s %*s %*s %*s %*s %*s %lu %lu", ib, ip, ob, op);
       active = 0;
       break;
     }
+  }
   fclose(fp);
 
   return active;
 }
 
-void
-linux_proc_term(struct Devices* dev)
-{}
-
 #endif /* USE_LINUX_PROC */
 
 
-/* FreeBDS sysctl driver */
+/* FreeBSD sysctl driver */
 #ifdef USE_FREEBSD_SYSCTL
 #undef drName
 #define drName USE_FREEBSD_SYSCTL
@@ -582,19 +564,20 @@ freebsd_sysctl_list(const char* devname, struct Devices* list)
   struct freebsd_sysctl_drvdata* drdata;
   int numifaces, numrfaces = 0;
   int mib[5], datamib[6];
-  int i, len, len2;
-  
+  int i;
+  size_t len, len2;
+
   mib[0] = CTL_NET;
   mib[1] = PF_LINK;
   mib[2] = NETLINK_GENERIC;
   mib[3] = IFMIB_SYSTEM;
   mib[4] = IFMIB_IFCOUNT;
-         
+
   datamib[0] = CTL_NET;
   datamib[1] = PF_LINK;
   datamib[2] = NETLINK_GENERIC;
   datamib[3] = IFMIB_IFDATA;
-  datamib[4] = 1; 
+  datamib[4] = 1;
   datamib[5] = IFDATA_GENERAL;
 
   len = sizeof(numifaces);
@@ -617,7 +600,7 @@ freebsd_sysctl_list(const char* devname, struct Devices* list)
     }
 
     if((devname && !strcmp(devname, tempndata.ifmd_name)) || (!devname &&
-        strcmp(tempndata.ifmd_name, "lo")))
+	strcmp(tempndata.ifmd_name, "lo")))
     {
       ndev = malloc(sizeof(struct Devices));
       ndev->devstart = 0;
@@ -641,14 +624,6 @@ freebsd_sysctl_list(const char* devname, struct Devices* list)
   return numrfaces;
 }
 
-/* per-interface initialization */
-int
-freebsd_sysctl_init(struct Devices* dev)
-{
-  /* FreeBSD doesn't need any device init */
-  return 0;
-}
-
 /* gather stats */
 int
 freebsd_sysctl_get(struct Devices*dev, unsigned long* ip,
@@ -656,7 +631,7 @@ freebsd_sysctl_get(struct Devices*dev, unsigned long* ip,
 {
   struct freebsd_sysctl_drvdata* drdata = dev->drvdata;
   int datamib[6];
-  int len;
+  size_t len;
 
   *ip = *op = *ib = *ob = 0;
 
@@ -667,7 +642,7 @@ freebsd_sysctl_get(struct Devices*dev, unsigned long* ip,
   datamib[4] = drdata->id;
   datamib[5] = IFDATA_GENERAL;
 
-  len = sizeof(struct ifmibdata); 
+  len = sizeof(struct ifmibdata);
 
   if(sysctl(datamib, 6, drdata->data, &len, NULL, 0) < 0)
     return 1;
@@ -727,7 +702,7 @@ irix_pcp_resDom(char* dom, pmID* pmId, pmDesc* pmD)
   if((r = pmLookupDesc(*pmId, pmD)) < 0)
   {
     msg_drInfo(drName, "unable to get descriptions about %s: %s",
-        dom, pmErrStr(r));
+	dom, pmErrStr(r));
     return -1;
   }
 
@@ -769,7 +744,7 @@ irix_pcp_list(const char* devname, struct Devices* list)
   if((dta = pmGetInDom(pmD->indom, &inst, &desc)) < 0)
   {
     msg_drInfo(drName, "unable to get instances of " PCPNS_NETINBDOM
-        ": %s", pmErrStr(dta));
+	": %s", pmErrStr(dta));
     return 0;
   }
 
@@ -779,7 +754,7 @@ irix_pcp_list(const char* devname, struct Devices* list)
   for(i = 0; i < dta; ++i)
   {
     if((devname && !strcmp(devname, *p)) ||
-        (!devname && strcmp(*p, "lo0")))
+	(!devname && strcmp(*p, "lo0")))
     {
       msg_drInfo(drName, "detected %s(%d)", *p, *t);
       ndev = (struct Devices*)malloc(sizeof(struct Devices));
@@ -836,7 +811,7 @@ irix_pcp_get(struct Devices* dev, unsigned long* ip, unsigned long* op,
   pmUseContext(drdata->ph);
   if(pmFetch(4, drdata->pmId, &pmR))
     return 1;
-  
+
   *ib = pmR->vset[0]->vlist->value.lval;
   *ip = pmR->vset[1]->vlist->value.lval;
   *ob = pmR->vset[2]->vlist->value.lval;
@@ -881,20 +856,6 @@ irix_pcp_term(struct Devices* dev)
 #define SNMPMIB_IFOUTBDEF "ifOutOctets"
 #define SNMPMIB_IFOUTPDEF "ifOutUcastPkts"
 
-/*
- * In WMND 0.4.5 I fetch ifName instead of ifDescr (ifName is shorter). I
- * noticed however that ifName is buggy on some firmware vendors (DLINK) and
- * there's no way to get around the bug without causing a network flood, so
- * simply invert the order.
- */
-#ifndef USE_GENERIC_SNMP_DESCR
-#define SNMPMIB_NAME_A "ifName"
-#define SNMPMIB_NAME_B "ifDescr"
-#else
-#define SNMPMIB_NAME_A "ifDescr"
-#define SNMPMIB_NAME_B "ifName"
-#endif
-
 struct generic_snmp_drvdata
 {
   struct snmp_session* se;
@@ -1035,8 +996,8 @@ generic_snmp_getDesc(struct snmp_session* se, int dev)
 {
   char* desc;
 
-  return ((desc = generic_snmp_getNodeDesc(se, SNMPMIB_NAME_A, dev))? desc:
-      generic_snmp_getNodeDesc(se, SNMPMIB_NAME_B, dev));
+  return ((desc = generic_snmp_getNodeDesc(se, "ifName", dev))? desc:
+      generic_snmp_getNodeDesc(se, "ifDescr", dev));
 }
 
 struct Devices*
@@ -1104,7 +1065,7 @@ generic_snmp_list(const char* devname, struct Devices* list)
     snmp_add_null_var(pdu, OID, OIDLen);
 
     if(snmp_synch_response(se, pdu, &res) == STAT_SUCCESS &&
-        res->errstat == SNMP_ERR_NOERROR)
+	res->errstat == SNMP_ERR_NOERROR)
       dta = *res->variables->val.integer;
     else
     {
@@ -1122,8 +1083,8 @@ generic_snmp_list(const char* devname, struct Devices* list)
       ndev = generic_snmp_preInit(list, se, rad + 1);
       if(!ndev)
       {
-        snmp_close(se);
-        continue;
+	snmp_close(se);
+	continue;
       }
 
       list = ndev;
@@ -1174,7 +1135,7 @@ generic_snmp_get(struct Devices* dev, unsigned long* ip, unsigned long* op,
   /* switch to the right context */
   struct generic_snmp_drvdata* drdata =
     (struct generic_snmp_drvdata*)dev->drvdata;
-  
+
   struct snmp_pdu* pdu;
   struct snmp_pdu* res;
   struct variable_list* var;
@@ -1186,7 +1147,7 @@ generic_snmp_get(struct Devices* dev, unsigned long* ip, unsigned long* op,
   snmp_add_null_var(pdu, drdata->oidPIn, drdata->oidPInLen);
   snmp_add_null_var(pdu, drdata->oidBOut, drdata->oidBOutLen);
   snmp_add_null_var(pdu, drdata->oidPOut, drdata->oidPOutLen);
-  
+
   if(snmp_synch_response(drdata->se, pdu, &res) == STAT_SUCCESS &&
       res->errstat == SNMP_ERR_NOERROR)
   {
@@ -1226,37 +1187,128 @@ generic_snmp_term(struct Devices* dev)
 
 #endif /* USE_GENERIC_SNMP */
 
+/* NetBSD ioctl driver */
+#ifdef USE_NETBSD_IOCTL
+#undef drName
+#define drName USE_NETBSD_IOCTL
+
+/* system headers */
+#include <sys/types.h>
+#include <sys/socket.h>
+#include <sys/sockio.h>
+#include <sys/ioctl.h>
+#include <net/if.h>
+#include <ifaddrs.h>
+
+int s;
+struct ifdatareq ifdreq;
+
+/* device listing */
+int
+netbsd_ioctl_list(const char *devname, struct Devices *list)
+{
+  struct Devices *ndev;
+  struct ifaddrs *ifap, *ifa;
+  unsigned int ifn;
+
+  if((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0)
+  {
+    msg_drInfo(drName, "unable to create socket");
+    return 0;
+  }
+
+  if(getifaddrs(&ifap) != 0)
+  {
+    msg_drInfo(drName, "failed to perform getifaddrs");
+    return 0;
+  }
+
+  ifn = 0;
+  for(ifa = ifap; ifa; ifa = ifa->ifa_next)
+  {
+    if(ifa->ifa_addr->sa_family == AF_LINK)
+    {
+      if((devname && !strcmp(devname, ifa->ifa_name)) || !devname)
+      {
+	ndev = NULL;
+	ndev = malloc(sizeof(struct Devices));
+	if(!ndev)
+	  continue;
+	ndev->devstart = 0;
+	ndev->name = strdup(ifa->ifa_name);
+	ndev->next = NULL;
+	list->next = ndev;
+	list = ndev;
+
+	msg_drInfo(drName, "detected %s", ndev->name);
+	ifn++;
+      }
+    }
+  }
+  freeifaddrs(ifap);
+
+  return ifn;
+}
+
+/* gather stats */
+int
+netbsd_ioctl_get(struct Devices *dev, unsigned long *ip,
+    unsigned long *op, unsigned long *ib, unsigned long *ob)
+{
+  strncpy(ifdreq.ifdr_name, dev->name, IFNAMSIZ - 1);
+  if(ioctl(s, SIOCGIFDATA, &ifdreq) < 0)
+    return 1;
+
+  *ip = ifdreq.ifdr_data.ifi_ipackets;
+  *ib = ifdreq.ifdr_data.ifi_ibytes;
+  *op = ifdreq.ifdr_data.ifi_opackets;
+  *ob = ifdreq.ifdr_data.ifi_obytes;
+
+  return 0;
+}
+
+void
+netbsd_ioctl_unlist()
+{
+  close(s);
+}
+
+#endif /* USE_NETBSD_IOCTL */
 
 /* define the drivers list */
 struct drivers_struct drivers_table[] =
 {
 #ifdef USE_FREEBSD_SYSCTL
-  {USE_FREEBSD_SYSCTL, freebsd_sysctl_list, freebsd_sysctl_init,
-    freebsd_sysctl_get, freebsd_sysctl_term},
+  {USE_FREEBSD_SYSCTL, freebsd_sysctl_list, NULL,
+    freebsd_sysctl_get, freebsd_sysctl_term, NULL},
+#endif
+#ifdef USE_NETBSD_IOCTL
+  {USE_NETBSD_IOCTL, netbsd_ioctl_list, NULL,
+    netbsd_ioctl_get, NULL, netbsd_ioctl_unlist},
 #endif
 #ifdef USE_LINUX_PROC
-  {USE_LINUX_PROC, linux_proc_list, linux_proc_init, linux_proc_get,
-    linux_proc_term},
+  {USE_LINUX_PROC, linux_proc_list, NULL,
+    linux_proc_get, NULL, NULL},
 #endif
 #ifdef USE_SOLARIS_FPPPD
   {USE_SOLARIS_FPPPD, solaris_fpppd_list, solaris_fpppd_init,
-    solaris_fpppd_get, solaris_fpppd_term},
+    solaris_fpppd_get, solaris_fpppd_term, NULL},
 #endif
 #ifdef USE_SOLARIS_KSTAT
   {USE_SOLARIS_KSTAT, solaris_kstat_list, solaris_kstat_init,
-    solaris_kstat_get, solaris_kstat_term},
+    solaris_kstat_get, solaris_kstat_term, NULL},
 #endif
 #ifdef USE_IRIX_PCP
-  {USE_IRIX_PCP, irix_pcp_list, irix_pcp_init, irix_pcp_get, irix_pcp_term},
+  {USE_IRIX_PCP, irix_pcp_list, irix_pcp_init,
+    irix_pcp_get, irix_pcp_term, NULL},
 #endif
 #ifdef USE_GENERIC_SNMP
   {USE_GENERIC_SNMP, generic_snmp_list, generic_snmp_init,
-    generic_snmp_get, generic_snmp_term},
+    generic_snmp_get, generic_snmp_term, NULL},
 #endif
 #ifdef USE_TESTING_DUMMY
   {USE_TESTING_DUMMY, testing_dummy_list, testing_dummy_init,
-    testing_dummy_get, testing_dummy_term},
+    testing_dummy_get, NULL, NULL},
 #endif
-  {NULL, NULL, NULL, NULL, NULL}
+  {NULL, NULL, NULL, NULL, NULL, NULL}
 };
-
