@@ -1,20 +1,11 @@
 /*
- * wmnd.c
- * by Reed Lai
+ * wmnd: main program module
  *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2 of the License, or
- * (at your option) any later version.
+ * Copyright(c) 2000-2001 by Reed Lai
+ * Copyright(c) 2001 by Timecop <timecop@japan.co.jp>
+ * Copyright(c) 2001-2008 by wave++ "Yuri D'Elia" <wavexx@users.sf.net>
  *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ * Distributed under GNU GPL (v2 or above) WITHOUT ANY WARRANTY.
  */
 
 #include "wmnd.h"
@@ -529,83 +520,84 @@ exec_perc_command(const char* cmd, int button)
 
 
 #ifdef USE_TREND
-const char*
-tmpfifo()
+/* static data for trend instance sharing */
+static FILE** trendFd;
+static int trendUpd;
+
+
+static inline int
+trend_idx(int d, int bp)
 {
-  const char* file;
-  int retries = 26;
-
-  while(retries--)
-  {
-    /* Nouuuuuu (sound of a lemming exploding)!
-     * Don't use THAT function! IT'S BAD! BAD BAAAD! Will kill 'ya!
-     * MAVAFF... I'd use tmpfile _anyway_ */
-    file = tmpnam(NULL);
-    if(!file) return NULL;
-    if(!mkfifo(file, 0600))
-      return file;
-  }
-
-  return NULL;
+  return(d * 2 + bp);
 }
 
 
 void
-exec_trend(int rt, int bp)
+close_trend(int i)
 {
-  int idx, i;
-  const char* title;
-  const char* file;
-  FILE* fd;
-
-  idx = (bp? 0: 2);	/* bytes or packets */
-  idx += (rt? 0: 1);	/* input or output */
-
-  /* title */
-  switch(idx)
+  if(trendFd[i])
   {
-  case 0: title = "input bytes"; break;
-  case 1: title = "ouput bytes"; break;
-  case 2: title = "input packets"; break;
-  case 3: title = "output packets"; break;
+    pclose(trendFd[i]);
+    trendFd[i] = NULL;
   }
+}
 
-  /* create the fifo */
-  file = tmpfifo();
-  if(!file)
+
+void
+close_trends()
+{
+  int i;
+  for(i = 0; i != wmnd.nr_devices * 2; ++i)
+    close_trend(i);
+  free(trendFd);
+}
+
+
+int
+check_trend(int i)
+{
+  return(fputs("\n", trendFd[i]) == EOF || fflush(trendFd[i]) == EOF);
+}
+
+
+void
+exec_trend(struct Devices* dev, int bp)
+{
+  int d = dev->devnum;
+  int i = trend_idx(d, bp);
+  int dead = (!trendFd[i] || check_trend(i));
+
+  if(dead)
   {
-    msg_err("cannot create temporary fifo");
-    return;
-  }
+    /* new trend instance required */
+    char cmd[256];
 
-  /* execute */
-  switch(fork())
-  {
-  case -1:
-    msg_err("cannot fork!");
-    break;
+    /* cmdline */
+    snprintf(cmd, sizeof(cmd),
+	"trend -t '%s %s' -G %d -d -s -c2a -Lin,out %s - %s",
+	dev->name, (bp? "bytes": "packets"),
+	(wmnd.scale == binary_scale? 1024: 1000) / (bp? 1: 100),
+	value("trend_flags"), value("trend_history"));
 
-  case 0:
-    execlp("trend", "trend", "-t", title, "-G", (bp? "1000": "10"),
-	"-d", "-m", file, "59", "58", NULL);
-
-    /* exec failed, unlock the parent anyway! */
-    msg_err("cannot execute trend!");
-    fd = fopen(file, "r");
-    exit(1);
-
-  default:
-    /* feed the pipe */
-    if((fd = fopen(file, "w")))
+    /* execute */
+    close_trend(i);
+    if(!(trendFd[i] = popen(cmd, "w")))
     {
-      for(i = 0; i != 59; ++i)
-	fprintf(fd, "%lu\n", wmnd.curdev->his[i][idx]);
-      fclose(fd);
+      msg_err("cannot execute trend!");
+      return;
     }
   }
 
-  /* remove the fifo */
-  unlink(file);
+  if(dead || !trendUpd)
+  {
+    /* refill the pipe */
+    int x, idx;
+    idx = (bp? 0: 2);
+    for(x = 0; x != 58; ++x)
+      fprintf(trendFd[i], "%lu %lu\n",
+	  dev->his[x][idx + 0], dev->his[x][idx + 1]);
+    fflush(trendFd[i]);
+  }
 }
 #endif
 
@@ -678,7 +670,7 @@ click_event(unsigned int region, unsigned int button)
 #ifdef USE_TREND
     case Button2:
       /* launch trend */
-      exec_trend((region == REG_SCALE_RX), bit_get(CFG_MODE));
+      exec_trend(wmnd.curdev, bit_get(CFG_MODE));
       break;
 #endif
     case Button3:
@@ -718,6 +710,9 @@ void
 mainExit(int sig)
 {
   msg_info("caught signal %d, terminating...", sig);
+#ifdef USE_TREND
+  close_trends();
+#endif
   devices_destroy();
   exit(1);
 }
@@ -740,11 +735,7 @@ int main(int argc, char* *argv)
   XEvent event;
   const struct drwStruct* drwPtr;
   sigset_t masked;
-#ifdef INEXACT_TIMING
-  int beats;
-#else
   struct timeval beat_time;
-#endif
 
   /* initialize messaging functions */
   msg_prgName = argv[0];
@@ -936,6 +927,13 @@ int main(int argc, char* *argv)
   sigaddset(&masked, SIGTERM);
   sigaddset(&masked, SIGINT);
 
+#ifdef USE_TREND
+  /* initialize trend's data */
+  signal(SIGPIPE, SIG_IGN);
+  trendFd = (FILE**)calloc(sizeof(FILE*), wmnd.nr_devices * 2);
+  trendUpd = strval_fe(psi_bool, value("trend_update"));
+#endif
+
   msg_dbg(__POSITION__, "open X display");
   dispname = value("display");
   if(!(dockapp.d = XOpenDisplay(dispname)))
@@ -958,11 +956,7 @@ int main(int argc, char* *argv)
   add_mr(REG_SCRIPT, 3, 54, 58, 7);	/* user script */
 
   /* updates should begin immediately */
-#ifdef INEXACT_TIMING
-  beats = wmnd.scroll * 100000 / wmnd.refresh;
-#else
   memset(&beat_time, 0, sizeof(beat_time));
-#endif
 
   /* clear the number of remaining steps */
   wmnd.avgRSteps = 1;
@@ -982,10 +976,9 @@ int main(int argc, char* *argv)
   {
     unsigned int j;
     sigset_t mask;
-#ifndef INEXACT_TIMING
     struct timeval beat_ctime;
     unsigned long beat_gap;
-#endif
+    int gap;
 
     /* mask INT/TERM in get_stats */
     sigprocmask(SIG_BLOCK, &masked, &mask);
@@ -1019,12 +1012,12 @@ int main(int argc, char* *argv)
       /* setup leds */
       if(ptr == wmnd.curdev)
       {
-	if(ptr->ip_stat_last == ip)
+	if(ptr->ip_stat_last == ip || ptr->online)
 	  led_control(LED_RX, 0);
 	else
 	  led_control(LED_RX, 1);
 
-	if(ptr->op_stat_last == op)
+	if(ptr->op_stat_last == op || ptr->online)
 	  led_control(LED_TX, 0);
 	else
 	  led_control(LED_TX, 1);
@@ -1050,31 +1043,18 @@ int main(int argc, char* *argv)
     /* restore mask */
     sigprocmask(SIG_SETMASK, &mask, NULL);
 
-#ifdef INEXACT_TIMING
-    /* cosmetic fix: use the number of effectively elapsed beats instead of a
-     * time based check to distribute timing discrepancies and rounding
-     * errors. This will also save us a syscall */
-    if((wmnd.refresh * beats / 100000) >= wmnd.scroll)
-    {
-      beats = 0;
-#define beat_gap wmnd.scroll
-#else
     /* fetch current time */
     gettimeofday(&beat_ctime, NULL);
 
-    /* estimate the time gap in tenth of seconds */
-    beat_gap = ((beat_ctime.tv_sec - beat_time.tv_sec) * 10) +
-      ((beat_ctime.tv_usec - beat_time.tv_usec) / 100000);
+    /* estimate the time gap in milliseconds */
+    beat_gap = ((beat_ctime.tv_sec - beat_time.tv_sec) * 1000) +
+      ((beat_ctime.tv_usec - beat_time.tv_usec) / 1000);
+    gap = (wmnd.avgSteps != 1? wmnd.scroll: beat_gap / 100);
 
-    if(beat_gap >= wmnd.scroll)
+    if(beat_gap >= wmnd.scroll * 100)
     {
       beat_time = beat_ctime;
 
-      /* Do not use the instant time gap for averaging, it's useless */
-      if(wmnd.avgSteps != 1)
-	beat_gap = wmnd.scroll;
-#endif
-
       /* drift the average stats */
       if(!--wmnd.avgRSteps)
       {
@@ -1098,9 +1078,33 @@ int main(int argc, char* *argv)
       beat_event();
 
       /* scroll the statistics */
-      draw_stats(wmnd.curdev, beat_gap);
+      draw_stats(wmnd.curdev, gap);
       for(ptr = devices; ptr; ptr = ptr->next)
       {
+#ifdef USE_TREND
+	if(trendUpd)
+	{
+	  int i, bp, idx;
+
+	  for(bp = 0; bp != 2; ++bp)
+	  {
+	    idx = (bp? 0: 2);
+	    i = trend_idx(ptr->devnum, bp);
+	    if(trendFd[i])
+	    {
+	      if(check_trend(i))
+		close_trend(i);
+	      else
+	      {
+		fprintf(trendFd[i], "%lu %lu\n",
+		    ptr->his[57][idx + 0], ptr->his[57][idx + 1]);
+		fflush(trendFd[i]);
+	      }
+	    }
+	  }
+	}
+#endif
+
 	for(j = 1; j < 58; j++)
 	{
 	  ptr->his[j - 1][0] = ptr->his[j][0];
@@ -1140,21 +1144,20 @@ int main(int argc, char* *argv)
 	  {
 	    click_event(rgn, btn);
 	    beat_event();
-	    draw_stats(wmnd.curdev, beat_gap);
+	    draw_stats(wmnd.curdev, gap);
 	  }
 	}
 	break;
       }
     }
 
-
-#ifdef INEXACT_TIMING
-    ++beats;
-#endif
     redraw_window();
     usleep(wmnd.refresh);
   }
 
+#ifdef USE_TREND
+  close_trends();
+#endif
   devices_destroy();
   exit(0);
 }
@@ -1190,7 +1193,7 @@ binary_scale(unsigned char sign, unsigned long value, char* buf)
   else
     scale = ' ';
 
-  sprintf(buf, "%c%lu", sign, value);
+  snprintf(buf, 7, "%c%lu", sign, value);
   r = buf;
   r++;
 
@@ -1237,7 +1240,7 @@ metric_scale(unsigned char sign, unsigned long value, char* buf)
   else
     scale = ' ';
 
-  sprintf(buf, "%c%f", sign, f);	/* to string */
+  snprintf(buf, 7, "%c%f", sign, f);
   r = buf;
   r++;
 
@@ -1392,8 +1395,8 @@ draw_string(const char* buf, unsigned int x, unsigned int y)
 void
 draw_rate(unsigned long rx, unsigned long tx, const int gap)
 {
-  char rx_buf[16];
-  char tx_buf[16];
+  char rx_buf[8];
+  char tx_buf[8];
 
   /* clear rate bar */
   copy_xpm_area(100, 85, 58, 7, 3, 54);
@@ -1636,9 +1639,7 @@ usage(void)
   int cnt = 0;
 
   fprintf(stderr,
-      "wmnd - WindowMaker Network Devices %s\n"
-      "Home page: http://www.yuv.info/wmnd/,\n"
-      "           http://wmnd.wingeer.org/\n\n"
+      "wmnd - WindowMaker Network Devices %s\n\n"
       "usage:\n"
       "  -b                  base 2 scale (no fractions)\n"
       "  -c <color>          tx color\n"
@@ -1743,6 +1744,7 @@ devices_init(const char* driver, const char* interface)
     for(in_loop0 = 0; in_loop0 < devnum; in_loop0++)
     {
       prt = prt->next;
+      prt->devnum = wmnd.nr_devices + in_loop0;
       prt->drvnum = cnt; /* set the driver number */
       if(drivers_table[cnt].init_device &&
 	  (*drivers_table[cnt].init_device)(prt) == 1) /* init the device */
